import argparse
import sys
from typing import Any
from typing import Dict
from typing import Union

import keyring
from jira import JIRA
from jira import JIRAError
from urllib3 import disable_warnings

from jira_select.types import InstanceDefinition

from ..constants import APP_NAME
from ..exceptions import UserError
from ..plugin import BaseCommand


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return (
            "Set up a Jira instance using command-line arguments.  "
            "Note: if you are a human, you probably want to use "
            "`configure` instead.  If `--password` is not specified "
            "as an argument, password will be read from stdin."
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--no-verify",
            "-n",
            help="Do not attempt to verify the provided credentials before saving.",
        )
        return super().add_arguments(parser)

    def handle(self, *args, **kwargs) -> None:
        verify: Union[bool, str] = True
        if self.options.disable_certificate_verification:
            verify = False
        elif self.options.certificate:
            verify = self.options.certificate

        password = self.options.password
        if not password:
            password = sys.stdin.read()

        if (
            not password
            or not self.options.instance_name
            or not self.options.username
            or not self.options.instance_url
        ):
            raise UserError(
                "Missing parameter; password, instance_name, username, "
                "and instance_url are all required."
            )

        if not self.options.no_verify:
            try:
                if self.options.disable_certificate_verification:
                    disable_warnings()

                options: Dict[str, Any] = {
                    "agile_rest_path": "agile",
                    "server": self.options.instance_url,
                    "verify": verify,
                }
                JIRA(
                    options,
                    basic_auth=(self.options.username, password),
                    max_retries=0,
                )
            except JIRAError as e:
                raise UserError(e)

        self.config.instances[self.options.instance_name] = InstanceDefinition(
            url=self.options.instance_url, username=self.options.username, verify=verify
        )
        self.save_config()

        keyring.set_password(
            APP_NAME,
            self.options.instance_url + self.options.username,
            password,
        )



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from pydantic import BaseModel
from pydantic import Extra
from pydantic import Field as ModelField

JiraFieldName = str


class SelectFieldDefinition(BaseModel):
    expression: str
    column: str


Field = Union[SelectFieldDefinition, JiraFieldName]

JQLString = str

CustomFilterFieldName = str

JqlList = List[str]
WhereParamDict = Dict[str, Any]

Expression = str

ExpressionList = List[Expression]


class QueryDefinition(BaseModel):
    select: Union[List[Field], Dict[str, Optional[str]]]
    static: Dict[str, str] = ModelField(default_factory=dict, alias="static")
    calculate: Dict[str, str] = ModelField(default_factory=dict)
    from_: str = ModelField(alias="from")
    subqueries: Dict[str, "QueryDefinition"] = ModelField(default_factory=dict)
    where: Union[JqlList, WhereParamDict] = ModelField(default_factory=list)
    order_by: JqlList = ModelField(default_factory=list)
    filter_: ExpressionList = ModelField(alias="filter", default_factory=list)
    having: ExpressionList = ModelField(default_factory=list)
    group_by: ExpressionList = ModelField(default_factory=list)
    sort_by: ExpressionList = ModelField(default_factory=list)
    expand: ExpressionList = ModelField(default_factory=list)
    limit: Optional[int]
    cap: Optional[int]
    cache: Optional[Union[int, Tuple[Optional[int], Optional[int]]]]

    class Config:
        allow_population_by_field_name = True
        extra = Extra.forbid


QueryDefinition.update_forward_refs()


class SchemaRow(BaseModel):
    id: str
    type: str
    description: Optional[str]
    raw: Optional[Any]


class ShellConfig(BaseModel):
    emacs_mode: Optional[bool] = False


class InstanceDefinition(BaseModel):
    url: Optional[str]
    username: Optional[str]
    password: Optional[str]
    verify: Optional[Union[str, bool]] = True


class ConfigDict(BaseModel):
    instances: Dict[str, InstanceDefinition] = ModelField(default_factory=dict)
    shell: ShellConfig = ModelField(default_factory=ShellConfig)
    inline_viewers: Dict[str, str] = ModelField(default_factory=dict)



from typing import Any
from typing import Optional

from dateutil.parser import parse as parse_datetime

from jira_select.plugin import BaseFunction


class Function(BaseFunction):
    """Parse a date string into a date object.

    This relies on `python-dateutil`; there are many additional options available
    that you can find documented `here <https://dateutil.readthedocs.io/en/stable/parser.html#dateutil.parser.parse>`_.

    """

    def __call__(self, date_str, *args, **kwargs) -> Optional[Any]:  # type: ignore[override]
        datetime = parse_datetime(date_str, *args, **kwargs)

        return datetime.date()



import argparse
import sys
from typing import IO

from prompt_toolkit.shortcuts import checkboxlist_dialog, input_dialog
from yaml import safe_dump

from ..exceptions import UserError
from ..plugin import BaseCommand


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return "Interactively generates a query definition (in yaml format)."

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--output",
            "-o",
            help="Path to file where records will be written; default: stdout.",
        )

    def get_select_fields(self):
        return checkboxlist_dialog(
            title="Select",
            text="Which fields would you like to select in your query?",
            values=[
                (
                    f'{field["id"]} as "{field["name"]}"',
                    f'{field["name"]} ({field["id"]})',
                )
                for field in sorted(self.jira.fields(), key=lambda x: x["name"])
            ],
        ).run()

    def get_where(self):
        return input_dialog(
            title="Where",
            text="(Optional) Enter a JQL query for limiting your search results.",
        ).run()

    def handle(self) -> None:
        fields = self.get_select_fields()
        if not fields:
            raise UserError("No fields selected")

        where = self.get_where()
        if not where:
            where = None

        output: IO[str] = sys.stdout
        if self.options.output:
            output = open(self.options.output, "w")

        query = {"select": fields, "from": "issues"}

        if where:
            query["where"] = [where]

        safe_dump(query, output)

        if output is not sys.stdout:
            output.close()



import json
from typing import Optional

from ..plugin import BaseFunction
from ..utils import JiraSelectJsonEncoder


class Function(BaseFunction):
    """Dump an object as a JSON string."""

    def __call__(self, value: str, *args, **kwargs) -> Optional[str]:  # type: ignore[override]
        return json.dumps(value, cls=JiraSelectJsonEncoder, *args, **kwargs)



import argparse
import subprocess
import sys
from typing import Optional
from typing import Tuple

from yaml import safe_load

from jira_select.exceptions import UserError
from jira_select.utils import launch_default_viewer

from ..constants import DEFAULT_INLINE_VIEWERS
from ..plugin import BaseCommand
from ..plugin import get_installed_formatters
from ..query import Executor
from ..types import QueryDefinition


def parameter_tuple(value: str) -> Tuple[str, str]:
    return tuple(value.split("=", 1))  # type: ignore


class Command(BaseCommand):
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        formatters = get_installed_formatters()

        parser.add_argument(
            "query_file",
            nargs="?",
            type=argparse.FileType("r"),
            default=sys.stdin,
            help="Query definition file to run",
        )
        parser.add_argument(
            "--format",
            "-f",
            choices=formatters.keys(),
            default="csv",
            help="Data format in which to write record data; default: 'csv'.",
        )
        parser.add_argument(
            "--output",
            "-o",
            type=argparse.FileType("wb+"),
            default=sys.stdout.buffer,
            help="Path to file where records will be written; default: stdout.",
        )
        parser.add_argument(
            "--launch-default-viewer",
            "-l",
            action="store_true",
            default=False,
            help=(
                "Display the output using the default viewer for the "
                "filetype used by the selected formatter instead of "
                "displaying the results inline."
            ),
            dest="launch_default_viewer",
        )
        parser.add_argument(
            "--view",
            "-v",
            default=False,
            action="store_true",
            help="Launch viewer immediately after completing query.",
        )
        parser.add_argument(
            "--param",
            "-p",
            dest="parameters",
            action="append",
            type=parameter_tuple,
        )
        parser.add_argument(
            "--no-cache",
            "-c",
            default=True,
            action="store_false",
            help="Do not use cached data.",
            dest="cache",
        )
        parser.add_argument(
            "--disable-progressbars",
            "-b",
            default=True,
            action="store_false",
            help="",
            dest="progressbar",
        )

    @classmethod
    def get_help(cls) -> str:
        return "Runs a query definition specified in yaml format."

    def handle(self) -> None:
        if self.options.view and self.options.output is sys.stdout:
            raise UserError("Must specify --output to use --view.")

        viewer: Optional[str] = self.config.inline_viewers.get(
            self.options.format, DEFAULT_INLINE_VIEWERS.get(self.options.format)
        )

        formatter_cls = get_installed_formatters()[self.options.format]

        query_definition: QueryDefinition
        query_definition = QueryDefinition.parse_obj(safe_load(self.options.query_file))

        query = Executor(
            self.jira,
            query_definition,
            progress_bar=(self.options.output is not sys.stdout.buffer)
            if self.options.progressbar
            else False,
            parameters=dict(self.options.parameters or []),
            enable_cache=self.options.cache,
        )
        with formatter_cls(query, self.options.output) as formatter:
            for row in query:
                formatter.writerow(row)
                self.options.output.flush()

        if self.options.view and self.options.launch_default_viewer:
            launch_default_viewer(self.options.output.name)
        elif self.options.view:
            if viewer:
                proc = subprocess.Popen([viewer, self.options.output.name])
                proc.wait()
            else:
                self.options.output.seek(0)
                self.console.print(self.options.output.read().decode("utf-8"))



from __future__ import annotations

from typing import Optional

import jira.resources

from jira_select.plugin import BaseFunction


class Function(BaseFunction):
    """Fetch a Jira issue by name."""

    def __call__(  # type: ignore[override]
        self,
        ticket_number: str,
    ) -> Optional[jira.resources.Issue]:
        if ticket_number:
            return self.jira.issue(ticket_number)

        return None



import datetime
from typing import Any
from typing import Iterable
from typing import Optional
from warnings import warn

from dateutil.tz import tzlocal
from pytz import UTC
from pytz import timezone

from jira_select.plugin import BaseFunction

from .flatten_changelog import flatten_changelog


class Function(BaseFunction):
    """Count the fractional number of work days an issue was in a given state."""

    def __call__(  # type: ignore[override]
        self,
        changelog: Any,
        state: str,
        start_hour: Optional[int] = 9,
        end_hour: Optional[int] = 17,
        timezone_name: Optional[str] = None,
        work_days: Iterable[int] = (1, 2, 3, 4, 5),
        min_date: datetime.date = datetime.date(1, 1, 1),
        max_date: datetime.date = datetime.date(9999, 1, 1),
    ) -> float:
        warn(
            "The `workdays_in_state` function is deprecated; see `interval_business_hours` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        tz = timezone(timezone_name) if timezone_name is not None else tzlocal()

        flattened_changelog = sorted(
            flatten_changelog(changelog),
            key=lambda row: row.created if row.created else datetime.date(1, 1, 1),
        )

        total_time = datetime.timedelta()

        cursor = min_date
        while cursor < max_date:
            if int(cursor.strftime("%w")) in work_days:
                min_day_date = datetime.datetime(
                    year=cursor.year,
                    month=cursor.month,
                    day=cursor.day,
                    hour=start_hour if start_hour is not None else 0,
                ).replace(tzinfo=tz)
                max_day_date = (
                    datetime.datetime(
                        year=cursor.year,
                        month=cursor.month,
                        day=cursor.day,
                        hour=end_hour,
                    ).replace(tzinfo=tz)
                    if end_hour is not None
                    else min_day_date + datetime.timedelta(days=1)
                )

                state_start: datetime.datetime | None = None
                for entry in flattened_changelog:
                    if entry.field != "status":
                        continue

                    if state_start is not None and entry.created:
                        valid_state_start = max(state_start, min_day_date)
                        valid_state_end = min(entry.created, max_day_date)
                        if valid_state_start < valid_state_end:
                            total_time += valid_state_end - valid_state_start
                        state_start = None
                    if entry.toString == state:
                        state_start = entry.created

                if state_start is not None:
                    valid_state_start = max(state_start, min_day_date)
                    valid_state_end = min(
                        max_day_date,
                        UTC.localize(datetime.datetime.utcnow()),
                    )
                    if valid_state_start < valid_state_end:
                        total_time += valid_state_end - valid_state_start

            cursor += datetime.timedelta(days=1)

        divisor = 1
        if end_hour is not None and start_hour is not None:
            divisor = 60 * 60 * (end_hour - start_hour)

        return total_time.total_seconds() / divisor



from typing import Any
from typing import Dict
from typing import List

from ..plugin import BaseFormatter


class Formatter(BaseFormatter):
    _fields: List[str]

    @classmethod
    def get_file_extension(cls) -> str:
        return "html"

    def open(self):
        super().open()

        self._fields = []

        for field in self.executor.query.select:
            self._fields.append(field.column)

        self.stream.write(
            b"""
            <!DOCTYPE html>
            <html>
                <head>
                    <style type="text/css">
                        body {
                            font-family: sans-serif;
                        }
                        table {
                            border-collapse: collapse;
                        }
                        td {
                            padding: 5px 10px;
                        }
                        tr {
                            border-bottom: 1px dotted black;
                        }
                    </style>
                </head>
                <body>
                    <table>
                        <thead>
                            <tr>
        """
        )
        for fieldname in self._fields:
            self.stream.write(f"<th>{fieldname}</th>".encode("utf-8"))

        self.stream.write(
            b"""
                </tr>
            </thead>
            <tbody>
        """
        )

    def close(self):
        self.stream.write(
            b"""
                        </tbody>
                    </table>
                </body>
            </html>
        """
        )
        return super().close()

    def writerow(self, row: Dict[str, Any]):
        self.stream.write(b"<tr>")
        for field in self._fields:
            self.stream.write(f"<td>{row.get(field)}</td>".encode("utf-8"))
        self.stream.write(b"</tr>")



from __future__ import annotations

import datetime
from types import SimpleNamespace
from unittest.mock import Mock

import pytz

from jira_select.functions.flatten_changelog import ChangelogEntry
from jira_select.functions.sprint_details import SprintInfo
from jira_select.plugin import get_installed_functions

from .base import JiraSelectTestCase


class JiraSelectFunctionTestCase(JiraSelectTestCase):
    def execute_function(self, name, *args, jira=None, **kwargs):
        if jira is None:
            jira = Mock()

        functions = get_installed_functions(jira)

        return functions[name](*args, **kwargs)


class TestSprintName(JiraSelectFunctionTestCase):
    def test_basic(self):
        sprint_string = "com.atlassian.greenhopper.service.sprint.Sprint@14b1c359[id=436,rapidViewId=153,state=CLOSED,name=MySprint,goal=Beep Boop,startDate=2020-03-09T21:53:07.264Z,endDate=2020-03-23T20:53:00.000Z,completeDate=2020-03-23T21:08:29.391Z,sequence=436]"

        actual = self.execute_function("sprint_name", sprint_string)
        expected = "MySprint"

        assert expected == actual


class TestSprintDetails(JiraSelectFunctionTestCase):
    def test_basic(self):
        sprint_string = "com.atlassian.greenhopper.service.sprint.Sprint@14b1c359[id=436,rapidViewId=153,state=CLOSED,name=MySprint,goal=Beep Boop,startDate=2020-03-09T21:53:07.264Z,endDate=2020-03-23T20:53:00.000Z,completeDate=2020-03-23T21:08:29.391Z,sequence=436]"

        expected = SprintInfo(
            id=436,
            rapidViewId=153,
            state="CLOSED",
            name="MySprint",
            goal="Beep Boop",
            startDate=datetime.datetime(2020, 3, 9, 21, 53, 7, 264000, tzinfo=pytz.UTC),
            endDate=datetime.datetime(2020, 3, 23, 20, 53, tzinfo=pytz.UTC),
            completeDate=datetime.datetime(
                2020, 3, 23, 21, 8, 29, 391000, tzinfo=pytz.UTC
            ),
            sequence=436,
        )
        actual = self.execute_function("sprint_details", sprint_string)

        assert expected == actual

    def test_extra_fields(self):
        sprint_string = "com.atlassian.greenhopper.service.sprint.Sprint@14b1c359[id=436,rapidViewId=153,state=CLOSED,name=MySprint,goal=Beep Boop,startDate=2020-03-09T21:53:07.264Z,endDate=2020-03-23T20:53:00.000Z,completeDate=2020-03-23T21:08:29.391Z,sequence=436,someUnexpectedField=UhOh]"

        expected = SprintInfo(
            id=436,
            rapidViewId=153,
            state="CLOSED",
            name="MySprint",
            goal="Beep Boop",
            startDate=datetime.datetime(2020, 3, 9, 21, 53, 7, 264000, tzinfo=pytz.UTC),
            endDate=datetime.datetime(2020, 3, 23, 20, 53, tzinfo=pytz.UTC),
            completeDate=datetime.datetime(
                2020, 3, 23, 21, 8, 29, 391000, tzinfo=pytz.UTC
            ),
            sequence=436,
        )
        actual = self.execute_function("sprint_details", sprint_string)

        assert expected == actual

    def test_missing_fields(self):
        sprint_string = "com.atlassian.greenhopper.service.sprint.Sprint@14b1c359[]"

        actual = self.execute_function("sprint_details", sprint_string)

        assert isinstance(actual, SprintInfo)


class TestGetSprint(JiraSelectFunctionTestCase):
    def test_basic(self):
        arbitrary_sprint_id = 10
        arbitrary_sprint_object = Mock(raw={})

        mock_jira = Mock(
            sprint=Mock(return_value=arbitrary_sprint_object),
            _options={"agile_rest_path": "/"},
        )

        self.execute_function("get_sprint_by_id", arbitrary_sprint_id, jira=mock_jira)

        assert mock_jira.sprint.called_with(arbitrary_sprint_id)


class TestFieldByName(JiraSelectFunctionTestCase):
    def test_basic(self):
        arbitrary_field_name = "customfield10000"
        arbitrary_human_name = "Sprint Name"
        arbitrary_value = "Beep Boop"

        mock_jira = Mock(
            fields=Mock(
                return_value=[
                    {"name": arbitrary_human_name, "key": arbitrary_field_name}
                ]
            )
        )

        my_row = SimpleNamespace(**{arbitrary_field_name: arbitrary_value})

        actual = self.execute_function(
            "field_by_name", my_row, arbitrary_human_name, jira=mock_jira
        )
        expected = arbitrary_value

        assert expected == actual


class TestEstimateToDays(JiraSelectFunctionTestCase):
    def test_basic(self):
        string_value = "1d 2h 3m"

        expected_value = 1.25625
        actual_value = self.execute_function("estimate_to_days", string_value)

        assert expected_value == actual_value

    def test_alternate_hour_day(self):
        string_value = "1d 2h 3m"

        expected_value = 1.5125
        actual_value = self.execute_function(
            "estimate_to_days", string_value, day_hour_count=4
        )

        assert expected_value == actual_value


class TestFlattenChangelog(JiraSelectFunctionTestCase):
    def test_basic(self):
        changelog = SimpleNamespace(
            histories=[
                SimpleNamespace(
                    **{
                        "author": "me@adamcoddington.net",
                        "created": "2015-11-01T18:16:37.388+0000",
                        "id": "10500",
                        "items": [
                            SimpleNamespace(
                                **{
                                    "field": "description",
                                    "fieldtype": "jira",
                                    "from": None,
                                    "fromString": None,
                                    "to": None,
                                    "toString": "Testing ticket body.",
                                }
                            )
                        ],
                    }
                ),
                SimpleNamespace(
                    **{
                        "author": "me@adamcoddington.net",
                        "created": "2015-10-11T07:39:40.063+0000",
                        "id": "10403",
                        "items": [
                            SimpleNamespace(
                                **{
                                    "field": "status",
                                    "fieldtype": "jira",
                                    "from": "10000",
                                    "fromString": "To Do",
                                    "to": "10001",
                                    "toString": "Done",
                                }
                            ),
                            SimpleNamespace(
                                **{
                                    "field": "resolution",
                                    "fieldtype": "jira",
                                    "from": None,
                                    "fromString": None,
                                    "to": "1",
                                    "toString": "Fixed",
                                }
                            ),
                        ],
                    }
                ),
            ]
        )

        expected_results = [
            ChangelogEntry(
                author="me@adamcoddington.net",
                created=datetime.datetime(
                    2015, 11, 1, 18, 16, 37, 388000, tzinfo=pytz.UTC
                ),
                id=10500,
                field="description",
                fieldtype="jira",
                fromValue=None,
                fromString=None,
                toValue=None,
                toString="Testing ticket body.",
            ),
            ChangelogEntry(
                author="me@adamcoddington.net",
                created=datetime.datetime(
                    2015, 10, 11, 7, 39, 40, 63000, tzinfo=pytz.UTC
                ),
                id=10403,
                field="status",
                fieldtype="jira",
                fromValue="10000",
                fromString="To Do",
                toValue="10001",
                toString="Done",
            ),
            ChangelogEntry(
                author="me@adamcoddington.net",
                created=datetime.datetime(
                    2015, 10, 11, 7, 39, 40, 63000, tzinfo=pytz.UTC
                ),
                id=10403,
                field="resolution",
                fieldtype="jira",
                fromValue=None,
                fromString=None,
                toValue="1",
                toString="Fixed",
            ),
        ]
        actual_results = self.execute_function("flatten_changelog", changelog)

        assert expected_results == actual_results


class TestSimpleFilter(JiraSelectFunctionTestCase):
    def test_basic(self):
        rows = [
            {
                "value": 10,
            },
            {
                "value": 20,
            },
        ]

        expected_results = [
            {
                "value": 10,
            }
        ]
        actual_results = self.execute_function("simple_filter", rows, value__lte=10)

        assert expected_results == actual_results

    def test_basic_multiple(self):
        rows = [{"value": 10, "name": "OK"}, {"value": 20, "name": "Nope"}]

        expected_results = [{"value": 10, "name": "OK"}]
        actual_results = self.execute_function(
            "simple_filter", rows, value__lte=99, name__eq="OK"
        )

        assert expected_results == actual_results


class TestSimpleFilterAny(JiraSelectFunctionTestCase):
    def test_basic(self):
        rows = [
            {
                "value": 10,
            },
            {
                "value": 20,
            },
        ]

        expected_results = [
            {
                "value": 10,
            }
        ]
        actual_results = self.execute_function("simple_filter_any", rows, value__lte=10)

        assert expected_results == actual_results

    def test_basic_multiple(self):
        rows = [{"value": 10, "name": "OK"}, {"value": 20, "name": "Nope"}]

        expected_results = [{"value": 10, "name": "OK"}, {"value": 20, "name": "Nope"}]
        actual_results = self.execute_function(
            "simple_filter_any", rows, value__lte=99, name__eq="OK"
        )

        assert expected_results == actual_results



from typing import Dict
from typing import Iterator
from typing import List

from jira.exceptions import JIRAError
from jira.resources import Sprint

from ..exceptions import QueryError
from ..plugin import BaseSource
from ..types import SchemaRow


class Source(BaseSource):
    SCHEMA: List[SchemaRow] = [
        SchemaRow.parse_obj({"id": "id", "type": "int"}),
        SchemaRow.parse_obj({"id": "state", "type": "str"}),
        SchemaRow.parse_obj({"id": "name", "type": "str"}),
        SchemaRow.parse_obj({"id": "startDate", "type": "datestr"}),
        SchemaRow.parse_obj({"id": "endDate", "type": "datestr"}),
        SchemaRow.parse_obj({"id": "completeDate", "type": "datestr"}),
        SchemaRow.parse_obj({"id": "originBoardId", "type": "int"}),
    ]

    def __iter__(self) -> Iterator[Dict]:
        start_at = 0
        max_results = 2**32
        result_limit = self.query.limit or 2**32

        if self.query.order_by:
            raise QueryError(
                "Sprint query 'order_by' expressions are not supported. "
                "Use 'sort_by' instead."
            )
        if self.query.expand:
            raise QueryError("Sprint query 'expand' expressions are not supported. ")

        where = self.query.where or {}
        if where and not isinstance(where, dict):
            raise QueryError(
                "Sprint query 'where' expressions should be a dictionary "
                "having any of the following keys: 'board_type', "
                "'board_name', or 'state'."
            )

        param_board_type = where.pop("board_type", None)
        param_board_name = where.pop("board_name", None)
        param_state = where.pop("state", None)

        if where:
            raise QueryError(f"Unexpected 'where' parameters: {where}.")

        count = 0

        already_returned = set()

        self.update_progress(completed=0, total=1, visible=True)
        while start_at < min(max_results, result_limit):
            results = self.jira.boards(
                startAt=start_at,
                maxResults=min(result_limit, 100),
                type=param_board_type,
                name=param_board_name,
            )

            max_results = results.total

            for result in results:
                self.update_progress(advance=1, total=results.total, visible=True)

                sprint_start_at = 0
                is_last = False
                while not is_last:
                    try:
                        sprints = self.jira.sprints(
                            result.id,
                            startAt=sprint_start_at,
                            maxResults=50,
                            state=param_state,
                        )
                    except JIRAError:
                        break

                    is_last = sprints.isLast
                    for sprint in sprints:
                        sprint_start_at += 1

                        if sprint.id in already_returned:
                            continue

                        count += 1
                        self.update_count(count)

                        yield sprint.raw

                        already_returned.add(sprint.id)

                        # Return early if our result limit has been reached
                        if count >= result_limit:
                            break

                start_at += 1

    def rehydrate(self, value: Dict) -> Sprint:
        return Sprint(
            {"agile_rest_path": self.jira._options["agile_rest_path"]}, None, value
        )



from typing import Dict
from typing import Optional

APP_NAME = "jira-select"


ENTRYPOINT_PREFIX = "jira_select"

COMMAND_ENTRYPOINT = f"{ENTRYPOINT_PREFIX}.commands"
SOURCE_ENTRYPOINT = f"{ENTRYPOINT_PREFIX}.sources"
FORMATTER_ENTRYPOINT = f"{ENTRYPOINT_PREFIX}.formatters"
FUNCTION_ENTRYPOINT = f"{ENTRYPOINT_PREFIX}.functions"


DEFAULT_INLINE_VIEWERS: Dict[str, Optional[str]] = {
    "csv": "vd",
    "json": "vd",
    "table": None,
}



from __future__ import annotations

import copy
from unittest.mock import ANY
from unittest.mock import Mock
from unittest.mock import patch

import simpleeval

from jira_select import query
from jira_select import utils
from jira_select.types import SelectFieldDefinition

from .base import JiraSelectTestCase


class TestParseSelectDefinition(JiraSelectTestCase):
    def test_returns_select_field_definition_directly(self):
        field_definition = SelectFieldDefinition.parse_obj(
            {
                "expression": "arbitrary",
                "column": "whatever",
            }
        )

        expected_result = copy.deepcopy(field_definition)
        actual_result = utils.parse_select_definition(field_definition)

        assert actual_result == expected_result

    def test_returns_name_if_no_as_expression(self):
        field_definition = "my_field_name"

        expected_result = SelectFieldDefinition.parse_obj(
            {
                "expression": field_definition,
                "column": field_definition,
            }
        )
        actual_result = utils.parse_select_definition(field_definition)

        assert expected_result == actual_result

    def test_returns_name_and_field_if_as_expression(self):
        field_definition = 'my_field_name as "My Field"'

        expected_result = SelectFieldDefinition.parse_obj(
            {
                "expression": "my_field_name",
                "column": "My Field",
            }
        )
        actual_result = utils.parse_select_definition(field_definition)

        assert expected_result == actual_result


class TestParseOrderByDefintition(JiraSelectTestCase):
    def test_handles_nonreversed(self):
        ordering = "somefield"

        expected_ordering = ordering, False
        actual_ordering = utils.parse_sort_by_definition(ordering)

        assert actual_ordering == expected_ordering

    def test_handles_reversed(self):
        ordering = "somefield desc"

        expected_ordering = "somefield", True
        actual_ordering = utils.parse_sort_by_definition(ordering)

        assert actual_ordering == expected_ordering

    def test_handles_reversed_casing(self):
        ordering = "somefield DESC"

        expected_ordering = "somefield", True
        actual_ordering = utils.parse_sort_by_definition(ordering)

        assert actual_ordering == expected_ordering

    def test_handles_nonreversed_casing(self):
        ordering = "somefield ASC"

        expected_ordering = "somefield", False
        actual_ordering = utils.parse_sort_by_definition(ordering)

        assert actual_ordering == expected_ordering


class TestCalculateResultHash(JiraSelectTestCase):
    def test_simple(self):
        issue = self.get_jira_issue()

        row = query.SingleResult(issue)
        assert isinstance(utils.calculate_result_hash(row, ["key"], {}), int)


class TestGetRowDict(JiraSelectTestCase):
    def test_copies_field_data(self):
        field_data = {
            "one": 1,
        }

        expected = {
            "key": ANY,
            "one": 1,
        }
        actual = utils.get_row_dict(self.get_jira_issue(field_data))

        assert actual == expected

    def test_copies_top_level_extra_keys(self):
        field_data = {
            "one": 1,
        }
        issue = self.get_jira_issue(field_data)
        issue.update(
            {
                "str_field": "ok",
                "float_field": 1.0,
                "int_field": 10,
                "list_field": [1, 2, 3],
                "dict_field": {"ok": True},
                "bool_field": False,
            }
        )

        expected = {
            "key": ANY,
            "str_field": "ok",
            "float_field": 1.0,
            "int_field": 10,
            "list_field": [1, 2, 3],
            "dict_field": {"ok": True},
            "bool_field": False,
            "one": 1,
        }
        actual = utils.get_row_dict(issue)

        assert actual == expected

    def test_does_not_copy_top_level_underscored(self):
        field_data = {
            "one": 1,
        }
        issue = self.get_jira_issue(field_data)
        issue.update({"_nope": 1})

        expected = {
            "key": ANY,
            "one": 1,
        }
        actual = utils.get_row_dict(issue)

        assert actual == expected


class TestEvaluateExpression(JiraSelectTestCase):
    def test_simple(self):
        expression = "my_field"
        names = {"my_field": "beep"}
        functions = {}

        expected_result = "beep"
        actual_result = utils.evaluate_expression(expression, names, functions)

        assert expected_result == actual_result

    def test_function(self):
        expression = "len(my_field)"
        names = {"my_field": "beep"}
        functions = {"len": len}

        expected_result = 4
        actual_result = utils.evaluate_expression(expression, names, functions)

        assert expected_result == actual_result

    def test_field_interpolations(self):
        arbitrary_interpolated_value = "boop"
        expression = "'{field name}'"
        names = {}
        interpolations = {"field name": arbitrary_interpolated_value}

        expected_result = arbitrary_interpolated_value
        actual_result = utils.evaluate_expression(
            expression, interpolations=interpolations, names=names
        )

        assert actual_result == expected_result


class TestGetFieldData(JiraSelectTestCase):
    def test_simple(self):
        mock_row = Mock(as_dict=Mock(return_value={"field": "OK"}))

        actual = utils.get_field_data(mock_row, "field")
        expected = "OK"

        assert actual == expected

    @patch("simpleeval.simple_eval")
    def test_none_for_dne(self, simple_eval):
        simple_eval.side_effect = simpleeval.AttributeDoesNotExist(None, None)

        mock_row = Mock(as_dict=Mock(return_value={"field": "OK"}))

        result = utils.get_field_data(mock_row, "arbitrary")
        assert result is None

    @patch("simpleeval.simple_eval")
    def test_none_for_keyerror(self, simple_eval):
        simple_eval.side_effect = KeyError()

        mock_row = Mock(as_dict=Mock(return_value={"field": "OK"}))

        result = utils.get_field_data(mock_row, "arbitrary")
        assert result is None

    @patch("simpleeval.simple_eval")
    def test_none_for_indexerror(self, simple_eval):
        simple_eval.side_effect = IndexError()

        mock_row = Mock(as_dict=Mock(return_value={"field": "OK"}))

        result = utils.get_field_data(mock_row, "arbitrary")
        assert result is None

    @patch("simpleeval.simple_eval")
    def test_none_for_typerror(self, simple_eval):
        simple_eval.side_effect = TypeError()

        mock_row = Mock(as_dict=Mock(return_value={"field": "OK"}))

        result = utils.get_field_data(mock_row, "arbitrary")
        assert result is None



from __future__ import annotations

import copy
import importlib.util
import json
import logging
import os
import random
import statistics
from abc import ABCMeta
from abc import abstractmethod
from typing import IO
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Type
from weakref import proxy

import keyring
import portion
from jira import JIRA
from rich.console import Console
from rich.progress import TaskID
from safdie import BaseCommand as SafdieBaseCommand
from safdie import get_entrypoints
from urllib3 import disable_warnings

from .constants import APP_NAME
from .constants import FORMATTER_ENTRYPOINT
from .constants import FUNCTION_ENTRYPOINT
from .constants import SOURCE_ENTRYPOINT
from .exceptions import ConfigurationError
from .types import ConfigDict
from .types import InstanceDefinition
from .types import SchemaRow
from .types import SelectFieldDefinition
from .utils import get_custom_function_dir
from .utils import get_functions_for_module
from .utils import save_config

if TYPE_CHECKING:
    from .query import CounterChannel
    from .query import Executor
    from .query import Query

logger = logging.getLogger(__name__)


BUILTIN_FUNCTIONS: Dict[str, Callable] = {
    # Built-ins
    "abs": abs,
    "all": all,
    "any": any,
    "bin": bin,
    "bool": bool,
    "hex": hex,
    "int": int,
    "len": len,
    "max": max,
    "min": min,
    "oct": oct,
    "ord": ord,
    "pow": pow,
    "range": range,
    "reversed": reversed,
    "round": round,
    "set": set,
    "sorted": sorted,
    "str": str,
    "sum": sum,
    "tuple": tuple,
    "map": map,
    "filter": filter,
    "type": lambda x: str(type(x)),
    # Statistics
    **get_functions_for_module(
        statistics,
        [
            "fmean",
            "geometric_mean",
            "harmonic_mean",
            "mean",
            "median",
            "median_grouped",
            "median_high",
            "median_low",
            "mode",
            "multimode",
            "pstdev",
            "pvariance",
            "quantiles",
            "stdev",
            "variance",
        ],
    ),
    # Random
    **get_functions_for_module(
        random,
        ["random", "randrange", "randint", "choice"],
    ),
    # JSON
    "json_loads": json.loads,
    "empty_interval": portion.empty,
    "closed_interval": portion.closed,
    "open_interval": portion.open,
    "openclosed_interval": portion.openclosed,
    "closedopen_interval": portion.closedopen,
}
REGISTERED_FUNCTIONS: Dict[str, Callable] = {}


class BaseCommand(SafdieBaseCommand):
    _jira: Optional[JIRA] = None

    def __init__(self, *, config: ConfigDict, **kwargs):
        self._config: ConfigDict = config
        self._console = Console(highlight=False)
        super().__init__(**kwargs)

    @property
    def config(self) -> ConfigDict:
        """Provides the configuration dictionary."""
        return self._config

    def save_config(self) -> None:
        """Saves the existing configuration dictionary."""
        save_config(self.config, self.options.config)

    @property
    def console(self) -> Console:
        """Provides access to the console (see `rich.console.Console`."""
        return self._console

    @property
    def jira(self) -> JIRA:
        """Provides access to the configured Jira instance."""
        if self._jira is None:
            instance = self.config.instances.get(
                self.options.instance_name, InstanceDefinition()
            )

            instance_url = self.options.instance_url or instance.url
            if not instance_url:
                raise ConfigurationError(
                    "instance_url not set; please run `jira-select configure`."
                )

            username = self.options.username or instance.username
            if not username:
                raise ConfigurationError(
                    "username not set; please run `jira-select configure`."
                )

            password = self.options.password or instance.password
            if not password:
                password = keyring.get_password(APP_NAME, instance_url + username)
                if not password:
                    raise ConfigurationError(
                        f"Password not stored for {instance_url} user {username}; "
                        "use the 'store-password' command to store the password "
                        "for this user account in your system keyring or use "
                        "`jira-select configure`."
                    )

            verify = self.options.disable_certificate_verification or instance.verify
            if verify is None:
                verify = True
            if verify is False:
                disable_warnings()

            self._jira = JIRA(
                options={
                    "agile_rest_path": "agile",
                    "server": instance_url,
                    "verify": verify,
                },
                basic_auth=(username, password),
                validate=False,
                get_server_info=False,
            )

        return self._jira


def get_installed_formatters() -> Dict[str, Type[BaseFormatter]]:
    return get_entrypoints(FORMATTER_ENTRYPOINT, BaseFormatter)


class BaseFormatter(metaclass=ABCMeta):
    def __init__(self, executor: Executor, stream: IO[bytes]):
        self._executor = proxy(executor)
        self._stream = stream

    @classmethod
    @abstractmethod
    def get_file_extension(cls) -> str:
        ...

    @property
    def executor(self):
        return self._executor

    @property
    def stream(self):
        return self._stream

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        # Exception handling here
        self.close()

    def open(self):
        return

    def close(self):
        return

    @abstractmethod
    def writerow(self, row: Dict[str, Any]):
        ...


def register_function(fn: Callable):
    """Register a callable to be a function available in queries."""
    REGISTERED_FUNCTIONS[fn.__name__] = fn


def get_installed_functions(
    jira: JIRA = None, executor: Executor = None
) -> Dict[str, Callable]:
    possible_commands: Dict[str, Callable] = copy.copy(BUILTIN_FUNCTIONS)

    # Import any modules in the custom functions directory; as a
    # side-effect of this, the functions will become listed within
    # REGISTERED_FUNCTIONS
    function_dir = get_custom_function_dir()
    for dirname, subdirlist, filelist in os.walk(function_dir):
        for filename in filelist:
            if filename.endswith(".py"):
                module_path_parts = ["user_scripts"]
                if dirname != function_dir:
                    module_path_parts.append(
                        dirname[len(function_dir) + 1 :]
                        .replace("/", ".")
                        .replace("\\", ".")
                    )
                module_path_parts.append(os.path.splitext(filename)[0])
                module_path = ".".join(module_path_parts)

                full_path = os.path.join(function_dir, dirname, filename)
                try:
                    spec = importlib.util.spec_from_file_location(
                        module_path, full_path
                    )
                    if not spec:
                        continue
                    user_module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(user_module)  # type: ignore
                except Exception as e:
                    logger.error("Could not import user script at %s: %s", full_path, e)

    possible_commands.update(REGISTERED_FUNCTIONS)

    for fn_name, fn in get_entrypoints(FUNCTION_ENTRYPOINT, BaseFunction).items():
        possible_commands[fn_name] = fn(jira, executor=executor)

    return possible_commands


class BaseFunction(metaclass=ABCMeta):
    def __init__(self, jira: Optional[JIRA], executor: Optional[Executor]):
        self._jira = jira
        self._executor = executor

    @property
    def query(self) -> Optional[Query]:
        return self.executor.query if self.executor else None

    @property
    def executor(self) -> Optional[Executor]:
        return self._executor

    @property
    def jira(self):
        assert self._jira

        return self._jira

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Optional[Any]:
        ...


def get_installed_sources() -> Dict[str, Type[BaseSource]]:
    return get_entrypoints(SOURCE_ENTRYPOINT, BaseSource)


class BaseSource(metaclass=ABCMeta):
    SCHEMA: List[SchemaRow] = []

    def __init__(self, executor: Executor, task: TaskID, out_channel: CounterChannel):
        self._executor = executor
        self._task = task
        self._out_channel = out_channel

        super().__init__()

    @classmethod
    def get_schema(cls, jira: JIRA) -> List[SchemaRow]:
        return copy.deepcopy(cls.SCHEMA)

    @classmethod
    def get_all_fields(cls, jira: JIRA) -> List[SelectFieldDefinition]:
        fields: List[SelectFieldDefinition] = []

        for entry in cls.get_schema(jira):
            fields.append(
                SelectFieldDefinition(
                    expression=entry.id,
                    column=entry.id,
                )
            )

        return fields

    def remove_progress(self):
        self._executor.progress.remove_task(self._task)

    def update_progress(self, *args, **kwargs):
        self._executor.progress.update(self._task, *args, **kwargs)

    def update_count(self, value: int):
        self._out_channel.set(value)

    @property
    def query(self) -> Query:
        return self._executor.query

    @property
    def jira(self) -> JIRA:
        return self._executor.jira

    @abstractmethod
    def rehydrate(self, value: Any) -> Any:
        ...

    @abstractmethod
    def __iter__(self) -> Iterator[Any]:
        ...



from __future__ import annotations

import operator
from functools import reduce
from typing import Iterable
from typing import TypeVar

from jira_select.plugin import BaseFunction

X = TypeVar("X")


class Function(BaseFunction):
    def __call__(self, iterable: Iterable[X]) -> X:  # type: ignore[override]
        return reduce(operator.or_, iterable)



from ..plugin import BaseCommand


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return (
            "Removes configuration for a particular Jira instance. "
            "Specify the instance to remove via `--instance-name`."
        )

    def handle(self, *args, **kwargs) -> None:
        del self.config.instances[self.options.instance_name]
        self.save_config()



from typing import Any
from typing import Dict

from ..plugin import BaseFormatter


class Formatter(BaseFormatter):
    @classmethod
    def get_file_extension(cls) -> str:
        return "raw"

    def writerow(self, row: Dict[str, Any]):
        for value in row.values():
            self.stream.write(str(value).encode("utf-8"))
            self.stream.write("\n".encode("utf-8"))



from __future__ import annotations

from abc import ABCMeta
from abc import abstractmethod
from functools import total_ordering
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

from dotmap import DotMap
from jira import JIRA
from pydantic import BaseModel
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TaskID
from rich.progress import TimeRemainingColumn

from . import __version__
from .cache import MinimumRecencyCache
from .exceptions import ExpressionParameterMissing
from .plugin import BaseSource
from .plugin import get_installed_functions
from .plugin import get_installed_sources
from .types import Expression
from .types import ExpressionList
from .types import Field
from .types import JqlList
from .types import QueryDefinition
from .types import SchemaRow
from .types import SelectFieldDefinition
from .types import WhereParamDict
from .utils import calculate_result_hash
from .utils import evaluate_expression
from .utils import expression_includes_group_by
from .utils import find_missing_parameters
from .utils import find_used_parameters
from .utils import get_cache_path
from .utils import get_field_data
from .utils import get_row_dict
from .utils import normalize_value
from .utils import parse_select_definition
from .utils import parse_sort_by_definition


@total_ordering
class NullAcceptableSort:
    def __init__(self, value: Any):
        self._value = value

    def __le__(self, other):
        if other._value is None:
            return True
        if self._value is None:
            return False

        return self._value < other._value

    def __eq__(self, other):
        if self._value is other._value:
            return True

        try:
            return self._value == other._value
        except Exception:
            return False

    def __str__(self):
        return str(self._value)


class Result(metaclass=ABCMeta):
    def __init__(self):
        self._overlay: dict[str, Any] = {}

    def __getattr__(self, name):
        result = self.as_dict()

        if name not in result:
            raise AttributeError(name)

        return result[name]

    @abstractmethod
    def as_dict(self) -> Dict[str, Any]:
        ...

    @abstractmethod
    def single(self) -> SingleResult:
        ...

    def evaluate_expression(
        self,
        expression: Expression,
        group_by: Optional[ExpressionList] = None,
        field_name_map: Optional[Dict[str, Any]] = None,
        functions: Optional[Dict[str, Callable]] = None,
    ):
        params: Dict[str, str] = cast(
            Dict[str, str],
            field_name_map.get("params", {}) if field_name_map is not None else {},
        )
        if missing := find_missing_parameters(
            expression,
            list(params.keys()),
        ):
            raise ExpressionParameterMissing(
                "Parameter {params.%s} found in expression, but no parameter was specified!"
                % missing[0]
            )

        if group_by is None:
            group_by = []

        row_source = self

        # Force returning a single value if the expression under
        # evaluation is one we're grouping on
        if expression_includes_group_by(expression, group_by):
            row_source = self.single()

        return get_field_data(
            row_source,
            expression,
            functions=functions,
            interpolations=field_name_map,
        )

    def __setitem__(self, name, value):
        self._overlay[name] = value


class SingleResult(Result):
    def __init__(self, row: Any):
        self._row = row
        super().__init__()

    def as_dict(self) -> Dict[str, Any]:
        return get_row_dict(self._row, self._overlay)

    def single(self) -> SingleResult:
        return self


class GroupedFieldContainer(list):
    def __getattr__(self, name):
        results = GroupedFieldContainer()

        for row in self:
            value = getattr(row, name, None)
            if value is not None:
                results.append(value)

        return results


class GroupedResult(Result):
    def __init__(
        self,
        rows: List[SingleResult] = None,
    ):
        super().__init__()
        self._rows: List[SingleResult] = rows if rows is not None else []

    @property
    def rows(self):
        return self._rows

    @property
    def all_fields(self):
        fields = set()

        for row in self._rows:
            fields |= set(row.as_dict())

        return fields

    def add(self, record: SingleResult):
        self._rows.append(record)

    def single(self) -> SingleResult:
        return self.rows[0]

    def as_dict(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        result.update(self._overlay)

        for field in self.all_fields:
            # Exclude empty rows -- in SQL when you run an aggregation
            # function on a set of rows, NULL rows are skipped, so we
            # should probably do the same?
            result[field] = GroupedFieldContainer(
                [
                    row.as_dict()[field]
                    for row in self._rows
                    if row.as_dict().get(field) is not None
                ]
            )

        return result


class CachedResults(BaseModel):
    source_schema: List[SchemaRow]
    rows: List[Dict]


class Query:
    def __init__(self, jira: JIRA, definition: QueryDefinition):
        self._jira = jira
        self._definition = definition

    def _ensure_str(self, iterable=Iterable[Any]) -> List[str]:
        return [str(item) for item in iterable]

    def _get_all_fields(self) -> List[SelectFieldDefinition]:
        sources = get_installed_sources()

        try:
            source = sources[self.from_]
        except KeyError:
            return []

        return source.get_all_fields(self._jira)

    def _get_select_calculate_fields(
        self, field_data: List[Field] | Mapping[str, str | None]
    ) -> List[SelectFieldDefinition]:
        fields: List[SelectFieldDefinition] = []

        if isinstance(field_data, str):
            field = field_data
            if field == "*":
                fields.extend(self._get_all_fields())
            else:
                fields.append(parse_select_definition(field))
        if isinstance(field_data, dict):
            for column, expression in field_data.items():
                fields.append(
                    SelectFieldDefinition(
                        expression=expression if expression else column, column=column
                    )
                )
        elif isinstance(field_data, list):
            for field in field_data:
                if field == "*":
                    fields.extend(self._get_all_fields())
                else:
                    fields.append(parse_select_definition(field))

        return fields

    @property
    def select(self) -> List[SelectFieldDefinition]:
        return self._get_select_calculate_fields(self._definition.select)

    @property
    def static(self) -> List[SelectFieldDefinition]:
        return self._get_select_calculate_fields(self._definition.static)

    @property
    def calculate(self) -> List[SelectFieldDefinition]:
        return self._get_select_calculate_fields(self._definition.calculate)

    @property
    def from_(self) -> str:
        return self._definition.from_

    @property
    def subqueries(self) -> dict[str, "QueryDefinition"]:
        return {
            subquery_name: QueryDefinition.parse_obj(subquery_value)
            for subquery_name, subquery_value in self._definition.subqueries.items()
        }

    @property
    def where(self) -> Union[JqlList, WhereParamDict]:
        where = self._definition.where
        if isinstance(where, list):
            return self._ensure_str(where)
        return where

    @property
    def order_by(self) -> JqlList:
        return self._ensure_str(self._definition.order_by)

    @property
    def filter(self) -> ExpressionList:
        return self._ensure_str(self._definition.filter_)

    @property
    def having(self) -> ExpressionList:
        return self._ensure_str(self._definition.having)

    @property
    def group_by(self) -> ExpressionList:
        return self._ensure_str(self._definition.group_by)

    @property
    def sort_by(self) -> List[Tuple[Expression, bool]]:
        return [
            parse_sort_by_definition(definition)
            for definition in self._ensure_str(self._definition.sort_by)
        ]

    @property
    def expand(self) -> List[str]:
        return self._ensure_str(self._definition.expand)

    @property
    def limit(self) -> Optional[int]:
        return self._definition.limit

    @property
    def cap(self) -> Optional[int]:
        return self._definition.cap

    @property
    def cache(self) -> Optional[Tuple[Optional[int], Optional[int]]]:
        value = self._definition.cache
        if isinstance(value, int):
            return (
                value,
                value,
            )
        return value


class NullProgressbar:
    def __init__(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def add_task(self, *args, **kwargs) -> TaskID:
        return TaskID(0)

    def remove_task(self, *args, **kwargs):
        pass

    def __enter__(self, *args, **kwargs) -> NullProgressbar:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        pass


class CounterChannel:
    def __init__(self):
        self._counter: int = 2**32

    def zero(self):
        self._counter = 0

    def increment(self):
        self._counter += 1

    def set(self, value: int):
        self._counter = value

    def get(self) -> int:
        return self._counter


class FieldNameMap(dict):
    def __missing__(self, key):
        return key


class Executor:
    def __init__(
        self,
        jira: JIRA,
        definition: QueryDefinition,
        enable_cache: bool = True,
        progress_bar: bool = False,
        parameters: Optional[Dict[str, Any]] = None,
        schema: Optional[List[SchemaRow]] = None,
    ):
        self._query: Query = Query(jira, definition)
        self._jira: JIRA = jira
        self._functions: Dict[str, Callable] = get_installed_functions(jira, self)
        self._progress_bar_enabled = progress_bar

        self._enable_cache = enable_cache
        self._cache = MinimumRecencyCache(get_cache_path())
        self._source_schema: List[SchemaRow] = schema if schema is not None else []
        self._field_name_map: Dict[str, str] = FieldNameMap()

        self._parameters: Dict[str, Any] = parameters or {}

    @property
    def jira(self) -> JIRA:
        return self._jira

    @property
    def schema(self) -> List[SchemaRow]:
        return self._source_schema

    @property
    def cache(self) -> MinimumRecencyCache:
        return self._cache

    @property
    def query(self) -> Query:
        return self._query

    @property
    def functions(self) -> Dict[str, Callable]:
        return self._functions

    @property
    def parameters(self) -> Dict[str, Any]:
        return self._parameters

    def get_source_schema(self) -> List[SchemaRow]:
        if not self._source_schema:
            sources = get_installed_sources()
            source = sources[self.query.from_]

            self._source_schema = source.get_schema(self.jira)

        return self._source_schema

    @property
    def field_name_map(self) -> Dict[str, Any]:
        if not self._field_name_map:
            for schema_row in self.get_source_schema():
                if not schema_row.description:
                    continue

                self._field_name_map[schema_row.description] = schema_row.id

            self._field_name_map["params"] = DotMap(self._parameters)

        return self._field_name_map

    def _get_cached(self, source: BaseSource) -> Iterator[Result]:
        cache_key = ":".join(
            [
                __version__,
                str(self.jira.client_info()),
                str(self.query.from_),
                str(self.query.where),
                str(self.query.order_by),
                str(self.query.limit),
                str(self.query.expand),
                str(
                    {
                        key: value
                        for key, value in self.parameters.items()
                        if key in find_used_parameters("".join(self.query.where))
                    }
                ),
            ]
        )

        if self.query.cache and self._enable_cache:
            try:
                min_recency, _ = self.query.cache
                if min_recency is None:
                    raise KeyError(cache_key)
                cached_results_raw = self.cache.get(cache_key, min_recency=min_recency)
                if not cached_results_raw:
                    raise KeyError(cache_key)

                cached_results = CachedResults.parse_obj(cached_results_raw)

                self._source_schema = cached_results.source_schema

                source.update_count(len(cached_results.rows))
                source.remove_progress()
                for result in cached_results.rows:
                    yield SingleResult(source.rehydrate(result))
                return
            except KeyError:
                pass

        cached_rows = []
        cached_schema = self.get_source_schema()

        for result in source:
            cached_rows.append(result)
            yield SingleResult(source.rehydrate(result))

        if self.query.cache:
            _, max_store = self.query.cache
            if max_store is not None:
                self.cache.set(
                    cache_key,
                    CachedResults(source_schema=cached_schema, rows=cached_rows).dict(),
                    expire=max_store,
                )

    def _get_static_results(
        self,
    ) -> Dict[str, Any]:
        shared: Dict[str, Any] = {}

        for definition in self.query.static:
            if missing := find_missing_parameters(
                definition.expression, list(self.parameters.keys())
            ):
                raise ExpressionParameterMissing(
                    "Parameter {params.%s} found in expression, but no parameter was specified!"
                    % missing[0]
                )

            shared[definition.column] = normalize_value(
                evaluate_expression(
                    definition.expression,
                    names={},
                    functions=self.functions,
                    interpolations={"params": DotMap(self.parameters)},
                )
            )

        return shared

    def _process_calculate(
        self,
        iterator: Iterator[Result],
        task: TaskID,
        input_channel: CounterChannel,
        output_channel: CounterChannel,
    ) -> Iterator[Result]:
        for row in iterator:
            output_channel.set(input_channel.get())
            self.progress.update(task, total=input_channel.get(), visible=True)

            for definition in self.query.calculate:
                row[definition.column] = self.evaluate_expression(
                    row, definition.expression
                )

            yield row

            self.progress.update(task, advance=1)

    def _process_filter(
        self,
        iterator: Iterator[Result],
        task: TaskID,
        input_channel: CounterChannel,
        output_channel: CounterChannel,
    ) -> Iterator[Result]:
        output_channel.zero()

        for row in iterator:
            self.progress.update(task, total=input_channel.get(), visible=True)

            include_row = True
            for filter_expression in self.query.filter:
                if not self.evaluate_expression(row, filter_expression):
                    include_row = False
                    break

            if include_row:
                output_channel.increment()
                yield row

            self.progress.update(task, advance=1)

    def _process_group_by(
        self,
        iterator: Iterator[Result],
        task: TaskID,
        input_channel: CounterChannel,
        output_channel: CounterChannel,
    ) -> Iterator[Result]:
        groups: Dict[int, Result] = {}

        output_channel.zero()

        for row in iterator:
            self.progress.update(task, total=input_channel.get(), visible=True)

            row_hash = calculate_result_hash(
                row,
                self.query.group_by,
                self.functions,
            )
            if row_hash not in groups:
                output_channel.increment()
                groups[row_hash] = GroupedResult()

            groups[row_hash].add(row)

            self.progress.update(task, advance=1)

        for _, value in groups.items():
            yield value

    def _process_having(
        self,
        iterator: Iterator[Result],
        task: TaskID,
        input_channel: CounterChannel,
        output_channel: CounterChannel,
    ) -> Iterator[Result]:
        output_channel.zero()

        for row in iterator:
            self.progress.update(task, total=input_channel.get(), visible=True)

            include_row = True
            for having in self.query.having:
                if not self.evaluate_expression(row, having):
                    include_row = False
                    break

            if include_row:
                output_channel.increment()
                yield row

            self.progress.update(task, advance=1)

    def _process_sort_by(
        self,
        iterator: Iterator[Result],
        task: TaskID,
        input_channel: CounterChannel,
        output_channel: CounterChannel,
    ) -> Iterator[Result]:
        # First, materialize our list
        rows = list(iterator)
        output_channel.set(len(rows))
        self.progress.update(task, total=len(rows), visible=True)

        # Now, sort by each of the ordering expressions in reverse order
        for sort_expression, reverse in reversed(self.query.sort_by):

            def sort_key(row):
                result = self.evaluate_expression(row, sort_expression)
                self.progress.update(task, advance=1)

                return NullAcceptableSort(result)

            rows = sorted(rows, key=sort_key, reverse=reverse)

        yield from rows

    def _generate_row_dict(self, row: Result) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        for field_defn in self.query.select:
            result[field_defn.column] = self.evaluate_expression(
                row,
                field_defn.expression,
            )

        return result

    def _get_iterator(self) -> Generator[Dict[str, Any], None, None]:
        sources = get_installed_sources()

        static_results = self._get_static_results()

        try:
            iterator = sources[self.query.from_]
        except KeyError:
            raise NotImplementedError(
                f"No search for source {self.query.from_} implemented."
            )

        # Do not show the counters until we know how many rows they have waiting
        iterator_task = self.progress.add_task("jira", total=0, visible=False)
        phases: List[
            Tuple[
                Callable[
                    [
                        Iterator[Result],
                        TaskID,
                        CounterChannel,
                        CounterChannel,
                    ],
                    Iterator[Result],
                ],
                TaskID,
            ]
        ] = []
        if self.query.calculate:
            calculate_task = self.progress.add_task("calculate", total=0, visible=False)
            phases.append((self._process_calculate, calculate_task))
        if self.query.filter:
            filter_task = self.progress.add_task("filter", total=0, visible=False)
            phases.append(
                (
                    self._process_filter,
                    filter_task,
                ),
            )
        if self.query.group_by:
            group_by_task = self.progress.add_task("group_by", total=0, visible=False)
            phases.append(
                (
                    self._process_group_by,
                    group_by_task,
                ),
            )
        if self.query.having:
            having_task = self.progress.add_task("having", total=0, visible=False)
            phases.append(
                (
                    self._process_having,
                    having_task,
                ),
            )
        if self.query.sort_by:
            sort_by_task = self.progress.add_task("sort_by", total=0, visible=False)
            phases.append(
                (
                    self._process_sort_by,
                    sort_by_task,
                ),
            )
        select_task = self.progress.add_task("select", total=0, visible=False)

        # Link up each generator with its source; these will vary
        # depending on what query feature are in use
        channel = CounterChannel()
        cursor: Iterator = self._get_cached(iterator(self, iterator_task, channel))
        for phase, task_id in phases:
            output_channel = CounterChannel()
            cursor = phase(cursor, task_id, channel, output_channel)
            channel = output_channel

        for row in cursor:
            self.progress.update(select_task, total=channel.get(), visible=True)

            for shared_result_name, shared_result_value in static_results.items():
                row[shared_result_name] = shared_result_value

            yield self._generate_row_dict(row)
            self.progress.update(select_task, advance=1)

    @property
    def progress(self) -> Union[Progress, NullProgressbar]:
        return self._progress_bar

    def evaluate_expression(self, row: Result, expression: Expression) -> Any:
        return row.evaluate_expression(
            expression,
            self.query.group_by,
            functions=self.functions,
            field_name_map=self.field_name_map,
        )

    def __iter__(self) -> Generator[Dict[str, Any], None, None]:
        progress_bar_cls: Union[Type[Progress], Type[NullProgressbar]] = NullProgressbar
        if self._progress_bar_enabled:
            progress_bar_cls = Progress

        with progress_bar_cls(
            "[progress.description]{task.description}",
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            "[medium_purple4]{task.completed}/{task.total}[/medium_purple4]",
            TimeRemainingColumn(),
        ) as progress:
            self._progress_bar = progress

            row_count = 0
            for row in self._get_iterator():
                yield row

                row_count += 1
                if self.query.cap is not None and row_count >= self.query.cap:
                    break



from unittest import TestCase
from unittest.mock import Mock
from typing import List, Dict, Optional, Any

from dotmap import DotMap

from jira_select.types import QueryDefinition
from jira_select.query import Executor


class JiraSelectTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self._issue_counter = 0

    def get_query(self, issues: List[Dict], query: QueryDefinition):
        return Executor(Mock(search_issues=Mock(return_value=issues)), query)

    def get_jira_issue(self, fields: Optional[Dict[str, Any]] = None):
        self._issue_counter += 1

        if fields is None:
            fields = {
                "int_field": 1,
                "bool_field": True,
                "list_field": [1, 2, 3],
                "dict_field": {"a": 1, "b": 2},
            }

        return DotMap({"key": f"ALPHA-{self._issue_counter}", "fields": fields})



from __future__ import annotations

import datetime
import hashlib
import json
import logging
import os
import re
import subprocess
import sys
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import simpleeval
from appdirs import user_config_dir
from jira.resources import Resource
from portion import Interval
from pytz import UTC
from simpleeval import EvalWithCompoundTypes
from yaml import safe_dump
from yaml import safe_load

from .constants import APP_NAME
from .exceptions import FieldNameError
from .exceptions import JiraSelectError
from .exceptions import QueryError
from .exceptions import UnhandledConditionError
from .types import ConfigDict
from .types import Expression
from .types import ExpressionList
from .types import SelectFieldDefinition

if TYPE_CHECKING:
    from .query import Result

FIELD_DISPLAY_DEFN_RE = re.compile(r'^(?P<expression>.*) as "(?P<column>.*)"$')
SORT_BY_DESC_FN = re.compile(r"^(?P<expression>.*) DESC", re.IGNORECASE)
SORT_BY_ASC_FN = re.compile(r"^(?P<expression>.*) ASC", re.IGNORECASE)
PARAM_FINDER = re.compile(r"{params\.([^}.]+)(?:.[^}.]+)?}")

ISO_FORMAT = "%Y-%m-%d %H:%M:%SZ"


logger = logging.getLogger(__name__)


class JiraSelectJsonEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        from .functions.flatten_changelog import ChangelogEntry

        if isinstance(obj, datetime.datetime):
            return UTC.normalize(obj).strftime(ISO_FORMAT)
        elif isinstance(obj, ChangelogEntry):
            return {
                "author": obj.author,
                "created": UTC.normalize(obj.created).strftime(ISO_FORMAT)
                if obj.created
                else None,
                "field": obj.field,
                "fieldtype": obj.fieldtype,
                "fromValue": obj.fromValue,
                "fromString": obj.fromString,
                "toValue": obj.toValue,
                "toString": obj.toString,
            }
        elif isinstance(obj, Interval):
            interval_list = []
            for interval in list(obj):
                interval_list.append(
                    (
                        interval.lower,
                        interval.upper,
                    )
                )
            return interval_list
        try:
            return super().default(obj)
        except TypeError:
            return str(obj)


def get_config_dir() -> str:
    root_path = user_config_dir(APP_NAME, "coddingtonbear")
    os.makedirs(root_path, exist_ok=True)

    return root_path


def get_custom_function_dir() -> str:
    root_path = os.path.join(user_config_dir(APP_NAME, "coddingtonbear"), "functions")
    os.makedirs(root_path, exist_ok=True)

    return root_path


def get_cache_path(subcache: str = "default") -> str:
    config_path = get_config_dir()
    cache_path = os.path.join(config_path, "cache", subcache)
    os.makedirs(cache_path, exist_ok=True)

    return cache_path


def get_default_config_path() -> str:
    root_path = get_config_dir()
    return os.path.join(
        root_path,
        "config.yaml",
    )


def get_config(path: str = None) -> ConfigDict:
    if path is None:
        path = get_default_config_path()

    if not os.path.isfile(path):
        return ConfigDict()

    with open(path, "r") as inf:
        return ConfigDict.parse_obj(safe_load(inf))


def save_config(data: ConfigDict, path: str = None) -> None:
    if path is None:
        path = get_default_config_path()

    with open(path, "w") as outf:
        safe_dump(data.dict(), outf)


def get_functions_for_module(
    module: ModuleType, names: List[str]
) -> Dict[str, Callable]:
    """Return functions in module matching the specified name.

    This exists because some functions we want to publish as available
    functions are available only on Python >= 3.8.

    """
    functions: Dict[str, Callable] = {}

    for name in names:
        if hasattr(module, name):
            functions[name] = getattr(module, name)

    return functions


def parse_select_definition(
    expression: Union[str, SelectFieldDefinition]
) -> SelectFieldDefinition:
    if isinstance(expression, str):
        as_match = FIELD_DISPLAY_DEFN_RE.match(expression)
        column = expression
        expression = expression
        if as_match:
            match_dict = as_match.groupdict()
            expression = match_dict["expression"]
            column = match_dict["column"]

        return SelectFieldDefinition(expression=expression, column=column)
    return expression


def parse_sort_by_definition(expression: str) -> Tuple[str, bool]:
    is_reversed = False
    is_reversed_match = SORT_BY_DESC_FN.match(expression)
    is_not_reversed_match = SORT_BY_ASC_FN.match(expression)
    if is_reversed_match:
        expression = is_reversed_match.groupdict()["expression"]
        is_reversed = True
    elif is_not_reversed_match:
        expression = is_not_reversed_match.groupdict()["expression"]
        is_reversed = False

    return expression, is_reversed


def calculate_result_hash(
    row: Result,
    group_fields: ExpressionList,
    functions: Dict[str, Callable],
) -> int:
    params = [
        str(get_field_data(row, group_field, functions)) for group_field in group_fields
    ]

    return int(hashlib.sha1(":".join(params).encode("UTF-8")).hexdigest(), 16)


def get_row_dict(row: Any, overlay: dict[str, Any] | None = None) -> Dict[str, Any]:
    names: dict[str, Any] = {}

    if hasattr(row, "fields"):
        for field_name in dir(row.fields):
            if not field_name.startswith("_"):
                names[field_name] = getattr(row.fields, field_name)

    # Gather any top-level keys, too, to make sure we fetch any expansions
    for key in dir(row):
        field_value = getattr(row, key)
        if (
            key != "fields"
            and not key.startswith("_")
            and not key.upper() == key
            and not callable(field_value)
        ):
            names[key] = getattr(row, key)

    names.update(overlay if overlay is not None else {})

    return names


def expression_includes_group_by(
    expression: Expression, group_by: List[Expression]
) -> bool:
    for group_by_expression in group_by:
        if group_by_expression in expression:
            return True

    return False


def evaluate_expression(
    expression: str,
    names: Dict[str, Any],
    functions: Optional[Dict[str, Callable]] = None,
    interpolations: Optional[Mapping[str, Any]] = None,
) -> Any:
    try:
        expression = expression.format_map(interpolations or {})
    except KeyError as e:
        raise FieldNameError(e)

    return EvalWithCompoundTypes(names=names, functions=functions).eval(expression)


def normalize_value(value: Any) -> Any:
    if isinstance(value, (str, bool, float, int)):
        return value
    elif isinstance(value, list):
        return [normalize_value(v) for v in value]
    elif isinstance(value, Dict):
        return {k: normalize_value(v) for k, v in value.items()}
    elif isinstance(value, Resource):
        display = str(value)
        if display.startswith("<JIRA"):
            # If the display value starts with "<JIRA", we know
            # the library couldn't find a pretty way of printing
            # this field; let's just return the raw dictionary
            # in that case
            return value.raw
        return display
    elif "PropertyHolder" in str(value):
        # Yes; this LOOKS ridiculous, but in older versions
        # of the Jira library, the PropertyHolder class is
        # created a runtime :facepalm:.  I think this is
        # the only way to find them :shrug:
        return {
            name: normalize_value(getattr(value, name))
            for name in dir(value)
            if not name.startswith("_") and not name.upper() == name
        }

    return value


def get_field_data(
    row: Result,
    expression: str,
    functions: Optional[Dict[str, Callable]] = None,
    interpolations: Optional[Mapping[str, Any]] = None,
    error_returns_null=True,
) -> Any:
    if functions is None:
        functions = {}

    try:
        return normalize_value(
            evaluate_expression(
                expression,
                names={
                    "_": row,  # Pre 3.0 queries
                    "issue": row,  # Post-3.0 queries
                    **row.as_dict(),
                },
                functions=functions,
                interpolations=interpolations,
            )
        )
    except FieldNameError as e:
        raise QueryError(f"Field {e} does not exist.") from e
    except JiraSelectError:
        # Re-raise these -- they can be used for alerting the user
        # to a configuration problem or w/e
        raise
    except (
        simpleeval.AttributeDoesNotExist,
        simpleeval.NameNotDefined,
        KeyError,
        IndexError,
        TypeError,
        AttributeError,
    ) as e:
        logger.warning(
            "%s while evaluating expression %s for issue(s) %s: %s",
            e.__class__.__name__,
            expression,
            row.key,
            e,
        )
        if not error_returns_null:
            raise
        return None
    except Exception as e:
        raise QueryError(f"{e}: {expression}") from e


def launch_default_viewer(relpath: str) -> None:
    path = os.path.abspath(relpath)

    if sys.platform == "win32":
        os.startfile(path)
    elif sys.platform == "darwin":
        subprocess.call(["open", path])
    elif sys.platform == "linux":
        subprocess.call(["xdg-open", path])
    else:
        raise UnhandledConditionError(
            "Could not determine method of launching default "
            f"interpreter for platform {sys.platform}."
        )


def find_used_parameters(expression: str) -> List[str]:
    used_params: Set[str] = set()

    expression_params = PARAM_FINDER.findall(expression)
    for expression_param in expression_params:
        used_params.add(expression_param)

    return list(used_params)


def find_missing_parameters(expression: str, known_params: List[str]) -> List[str]:
    missing_parameters: List[str] = []

    expression_params = PARAM_FINDER.findall(expression)
    for expression_param in expression_params:
        if expression_param not in known_params:
            missing_parameters.append(expression_param)

    return missing_parameters



from __future__ import annotations

import datetime
from typing import Any

import portion

from jira_select.plugin import BaseFunction


class Function(BaseFunction):
    def __call__(  # type: ignore[override]
        self, interval: portion.Interval, default=datetime.timedelta(seconds=0)
    ) -> portion.Interval:
        total: Any = default

        for subinterval in interval:
            result = subinterval.upper - subinterval.lower
            if total is None:
                total = result
            else:
                total += result

        return total



from typing import Dict
from typing import Hashable
from typing import Iterator
from typing import Optional
from typing import Union

from jira.exceptions import JIRAError
from jira.resources import Sprint

from ..exceptions import UserError
from ..plugin import BaseFunction


class Function(BaseFunction):
    """Returns a `jira.resources.Sprint` object for a given sprint ID string."""

    CACHE: Dict[Hashable, Sprint] = {}

    def get_sprint_iterator(
        self, board_name_or_id: Union[str, int]
    ) -> Iterator[Sprint]:
        if isinstance(board_name_or_id, int):
            is_last = False
            start_at = 0

            while not is_last:
                try:
                    sprints = self.jira.sprints(
                        board_name_or_id, startAt=start_at, maxResults=50
                    )
                except JIRAError:
                    break

                is_last = sprints.isLast
                for sprint in sprints:
                    yield sprint
                    start_at += 1
        else:
            board_start = 0
            board_is_last = False

            while not board_is_last:
                boards = self.jira.boards(
                    startAt=board_start, maxResults=50, name=board_name_or_id
                )

                board_is_last = boards.isLast
                for board in boards:
                    is_last = False
                    start_at = 0

                    while not is_last:
                        try:
                            sprints = self.jira.sprints(
                                board.id, startAt=start_at, maxResults=50
                            )
                        except JIRAError:
                            break

                        is_last = sprints.isLast
                        for sprint in sprints:
                            yield sprint
                            start_at += 1

                    board_start += 1

    def get_sprint_details(
        self, board_name_or_id: Union[str, int], sprint_name: str
    ) -> Sprint:
        cache_key = (
            board_name_or_id,
            sprint_name,
        )

        if cache_key not in self.CACHE:
            sprints = list(
                filter(
                    lambda sprint: sprint_name.lower() in sprint.name.lower(),
                    self.get_sprint_iterator(board_name_or_id),
                )
            )

            if len(sprints) == 0:
                raise UserError(
                    f"No sprint found on board {board_name_or_id} "
                    f"named {sprint_name}."
                )
            elif len(sprints) > 1:
                raise UserError(
                    f"More than one sprint was found on {board_name_or_id} "
                    f"having a name like {sprint_name}: {sprints}."
                )

            self.CACHE[cache_key] = Sprint(
                {"agile_rest_path": self.jira._options["agile_rest_path"]},
                None,
                sprints[0].raw,
            )

        return self.CACHE[cache_key]

    def __call__(self, board_name_or_id: Union[str, int], sprint_name: str) -> Optional[Sprint]:  # type: ignore[override]
        if board_name_or_id is None or sprint_name is None:
            return None

        return self.get_sprint_details(board_name_or_id, sprint_name)



from __future__ import annotations

import datetime
from typing import Iterable

import portion
from dateutil.tz import tzlocal
from pytz import timezone

from jira_select.plugin import BaseFunction


def get_business_hours_intervals(
    start: datetime.date,
    end: datetime.date,
    start_hour: int = 9,
    end_hour: int = 17,
    timezone_name: str | None = None,
    work_days: Iterable[int] = (1, 2, 3, 4, 5),
) -> portion.Interval:
    business_hours = portion.empty()

    tz = timezone(timezone_name) if timezone_name is not None else tzlocal()
    cursor = start
    while cursor < end:
        if int(cursor.strftime("%w")) in work_days:
            min_day_date = datetime.datetime(
                year=cursor.year,
                month=cursor.month,
                day=cursor.day,
                hour=start_hour,
            ).replace(tzinfo=tz)
            max_day_date = datetime.datetime(
                year=cursor.year,
                month=cursor.month,
                day=cursor.day,
                hour=end_hour,
            ).replace(tzinfo=tz)

            business_hours |= portion.closed(min_day_date, max_day_date)

        cursor += datetime.timedelta(days=1)

    return business_hours


class Function(BaseFunction):
    def __call__(  # type: ignore[override]
        self,
        min_date: datetime.date | None = None,
        max_date: datetime.date | None = None,
        start_hour: int = 9,
        end_hour: int = 17,
        timezone_name: str | None = None,
        work_days: Iterable[int] = (1, 2, 3, 4, 5),
    ) -> portion.Interval:
        if min_date is None:
            min_date = datetime.datetime.now().date() - datetime.timedelta(days=365)
        if max_date is None:
            max_date = datetime.datetime.now().date() + datetime.timedelta(days=1)

        return get_business_hours_intervals(
            min_date,
            max_date,
            start_hour,
            end_hour,
            timezone_name,
            work_days,
        )



from typing import Any, List

from ..plugin import BaseFunction


class Function(BaseFunction):
    def __call__(self, scalar: List[List[Any]]) -> Any:  # type: ignore[override]
        return [item for sublist in scalar if sublist for item in sublist if item]



from logging import getLogger
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union

import keyring
from jira import JIRA
from jira import JIRAError
from prompt_toolkit.shortcuts import input_dialog
from prompt_toolkit.shortcuts import yes_no_dialog
from urllib3 import disable_warnings

from jira_select.types import InstanceDefinition

from ..constants import APP_NAME
from ..exceptions import UserError
from ..plugin import BaseCommand

logger = getLogger(__name__)


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return (
            "Interactively allows you to configure jira-select to connect "
            "to your Jira instance."
        )

    def collect_credentials(self) -> Tuple[str, str, str]:
        instance_url = input_dialog(
            title="Instance URL",
            text="Please enter your Jira instance URL (e.g. 'https://mycompany.jira.com/'):",
        ).run()
        if not instance_url:
            raise UserError("Cancelled")
        username = input_dialog(
            title="Username",
            text=(
                "Please enter your Jira username: \n\n(for Jira Cloud instances, you may need to generate a new API token to use as a password at https://id.atlassian.com/manage-profile/security/api-tokens)"
            ),
        ).run()
        if not username:
            raise UserError("Cancelled")
        password = input_dialog(
            title="Password",
            text=f"Please enter the password for {username}: ",
            password=True,
        ).run()
        if not password:
            raise UserError("Cancelled")

        return instance_url, username, password

    def handle(self) -> None:
        instance_url = ""
        username = ""
        password = ""
        verify: Union[bool, str] = True

        while True:
            instance_url, username, password = self.collect_credentials()

            if self.options.disable_certificate_verification:
                verify = False
                disable_warnings()
            elif self.options.certificate:
                verify = self.options.certificate

            options: Dict[str, Any] = {
                "agile_rest_path": "agile",
                "server": instance_url,
                "verify": verify,
            }

            try:
                JIRA(
                    options,
                    basic_auth=(username, password),
                    max_retries=0,
                )
                break
            except JIRAError:
                result = yes_no_dialog(
                    title="Error connecting to Jira",
                    text=(
                        "A connection to Jira could not be established using "
                        "the credentials you provided; try again?"
                    ),
                ).run()
                if not result:
                    raise UserError("Aborted; configuration not saved.")

        self.config.instances[self.options.instance_name] = InstanceDefinition(
            url=instance_url,
            username=username,
            verify=verify,
        )
        self.save_config()

        store_password = yes_no_dialog(
            title="Save password?",
            text="Would you like to save this password to your system keyring?",
        ).run()
        if store_password:
            keyring.set_password(APP_NAME, instance_url + username, password)



import argparse
import inspect
import json

from rich.table import Table

from ..plugin import BaseCommand
from ..plugin import BaseFunction
from ..plugin import get_installed_functions
from ..utils import JiraSelectJsonEncoder
from ..utils import evaluate_expression


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return "Lists all available functions, their signatures, and description."

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "search_terms",
            nargs="*",
            type=str,
            help="Case-insensitive search term for limiting displayed results.",
        )
        parser.add_argument(
            "--json",
            "-j",
            default=False,
            action="store_true",
            help="Export functions in JSON format instead of printing a table.",
        )
        parser.add_argument(
            "--having",
            help=(
                "A 'having' expression to use for limiting the displayed results. "
                "Searchable fields are 'name' and 'description'. "
                "E.g.--having=\"'average' in description.lower()\". "
            ),
        )

    def get_dotpath(self, fn) -> str:
        if isinstance(fn, BaseFunction):
            return f"{fn.__class__.__module__}.{fn.__class__.__name__}"
        return f"{fn.__module__}.{fn.__name__}"

    def handle(self):
        functions = get_installed_functions(self.jira)

        if self.options.json:
            function_data = []
            for name, fn in functions.items():
                docstring = fn.__doc__ or ""

                row = {
                    "name": name,
                    "description": docstring,
                    "dotpath": self.get_dotpath(fn),
                }
                try:
                    row["signature"] = str(inspect.signature(fn))
                except Exception:
                    pass

                if self.options.search_terms:
                    matches = True
                    for option in self.options.search_terms:
                        if option.lower() not in str(row).lower():
                            matches = False
                            break
                    if not matches:
                        continue
                if self.options.having:
                    if not evaluate_expression(self.options.having, row, functions):
                        continue

                function_data.append(row)
            function_data.sort(key=lambda row: row["name"])
            self.console.print(
                json.dumps(
                    function_data, cls=JiraSelectJsonEncoder, sort_keys=True, indent=4
                )
            )
        else:
            table = Table(title="functions")
            table.add_column(header="name", style="green")
            table.add_column(header="description", style="white")

            all_functions = []
            for name, fn in functions.items():
                docstring = fn.__doc__ or ""

                try:
                    signature = (
                        f"[bright_cyan]{name}{inspect.signature(fn)}[/bright_cyan]\n"
                    )
                except Exception:
                    signature = ""

                dotpath = f"[blue]{self.get_dotpath(fn)}[/blue]\n\n"

                description = signature + dotpath + docstring.strip() + "\n\n"

                row = {
                    "name": name,
                    "description": description,
                }
                if self.options.search_terms:
                    matches = True
                    for option in self.options.search_terms:
                        if option.lower() not in str(row).lower():
                            matches = False
                            break
                    if not matches:
                        continue
                if self.options.having:
                    if not evaluate_expression(self.options.having, row, functions):
                        continue

                all_functions.append(row)

            all_functions.sort(key=lambda row: row["name"])
            for row in all_functions:
                table.add_row(row["name"], row["description"])

            self.console.print(table)



import argparse
import json
from typing import Any
from typing import Callable
from typing import Dict

from rich.table import Table

from ..exceptions import JiraSelectError
from ..plugin import BaseCommand
from ..plugin import get_installed_functions
from ..plugin import get_installed_sources
from ..utils import JiraSelectJsonEncoder
from ..utils import evaluate_expression


class Command(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.functions: Dict[str, Callable] = get_installed_functions(self.jira)

    @classmethod
    def get_help(cls) -> str:
        return (
            "Describes the fields available in data returned from "
            "a particular data source."
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        sources = get_installed_sources()

        parser.add_argument(
            "source",
            choices=list(sources.keys()),
        )
        parser.add_argument(
            "search_terms",
            nargs="*",
            type=str,
            help="Case-insensitive search term for limiting displayed results.",
        )
        parser.add_argument(
            "--json",
            "-j",
            default=False,
            action="store_true",
            help="Export schema as JSON format instead of printing a table.",
        )
        parser.add_argument(
            "--having",
            help=(
                "A 'having' expression to use for limiting the displayed results. "
                "Searchable fields are 'key', 'type', and 'description' and 'raw'. "
                "E.g.--having=\"'estimate' in description.lower()\". "
            ),
        )

    def evaluate_expression(self, row: Any, expression: str) -> Any:
        return evaluate_expression(expression, row, functions=self.functions)

    def handle(self) -> None:
        sources = get_installed_sources()

        try:
            source = sources[self.options.source]
        except KeyError:
            raise JiraSelectError(f"No source named '{self.options.source}' found.")

        if self.options.json:
            fields = []
            for row in source.get_schema(self.jira):
                if self.options.search_terms:
                    matches = True
                    for option in self.options.search_terms:
                        if option.lower() not in str(row).lower():
                            matches = False
                            break
                    if not matches:
                        continue
                if self.options.having:
                    if not self.evaluate_expression(row, self.options.having):
                        continue

                fields.append(
                    {
                        "id": row.id,
                        "type": row.type,
                        "description": row.description,
                    }
                )
            self.console.print(
                json.dumps(fields, cls=JiraSelectJsonEncoder, sort_keys=True, indent=4)
            )
        else:
            table = Table(title=self.options.source)
            table.add_column(header="id", style="green")
            table.add_column(header="type", style="cyan")
            table.add_column(header="description", style="bright_cyan")

            for row in source.get_schema(self.jira):
                if self.options.search_terms:
                    matches = True
                    for option in self.options.search_terms:
                        if option.lower() not in str(row).lower():
                            matches = False
                            break
                    if not matches:
                        continue
                if self.options.having:
                    if not self.evaluate_expression(row, self.options.having):
                        continue

                table.add_row(row.id, row.type, row.description)

            self.console.print(table)



class JiraSelectError(Exception):
    pass


class ConfigurationError(JiraSelectError):
    pass


class UnhandledConditionError(JiraSelectError):
    pass


class UserError(JiraSelectError):
    pass


class QueryError(UserError):
    pass


class FieldNameError(QueryError):
    pass


class ExpressionParameterMissing(QueryError):
    pass
